"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AppBase = exports.DEFAULT_RATE_LIMIT_MAX_REQ_PER_WINDOW = exports.DEFAULT_RATE_LIMIT_WINDOW = exports.express = void 0;
const body_parser_1 = __importDefault(require("body-parser"));
const cors_1 = __importDefault(require("cors"));
const express_1 = __importDefault(require("express"));
exports.express = express_1.default;
const express_rate_limit_1 = __importDefault(require("express-rate-limit"));
const helmet_1 = __importDefault(require("helmet"));
const multer_1 = __importDefault(require("multer"));
const cm_libs_context_aware_error_1 = require("cm-libs-context-aware-error");
const cm_libs_context_aware_logger_1 = require("cm-libs-context-aware-logger");
const cm_libs_utils_1 = require("cm-libs-utils");
const common_1 = require("./common");
exports.DEFAULT_RATE_LIMIT_WINDOW = 15 * 60000; // 15 minutes
exports.DEFAULT_RATE_LIMIT_MAX_REQ_PER_WINDOW = 100;
class AppBase {
    constructor(name, version) {
        this._devMode = false;
        this._terminated = false;
        this.name = name || '';
        this.version = version || '';
    }
    async init(origins, rateLimitHandler, overrides) {
        // Setup bootstrap before proceed
        this.bootstrap();
        this.app = (0, express_1.default)();
        this.app.set('x-powered-by', false);
        if (!this.devMode) {
            this.app.use((0, helmet_1.default)());
        }
        // Apply the rate limiting middleware to all requests
        if (rateLimitHandler) {
            this.app.use((0, express_rate_limit_1.default)({
                windowMs: 0 < overrides?.RATE_LIMIT_WINDOW ? overrides?.RATE_LIMIT_WINDOW : exports.DEFAULT_RATE_LIMIT_WINDOW,
                max: 0 < overrides?.RATE_LIMIT_MAX_REQ_PER_WINDOW ? overrides?.RATE_LIMIT_MAX_REQ_PER_WINDOW : exports.DEFAULT_RATE_LIMIT_MAX_REQ_PER_WINDOW,
                standardHeaders: true,
                legacyHeaders: false,
                handler: rateLimitHandler
            }));
        }
        // Setup default CORS
        this.app.options('*', (0, cors_1.default)());
        this.app.use((0, cors_1.default)({
            origin: origins?.split(',') || ['*'],
            allowedHeaders: 'Origin,X-Requested-With,Content-Type,Accept,Lang,Authorization,Accept-Version,Content-Language,app-id,Location',
            exposedHeaders: 'X-Requested-With,X-Metadata',
            preflightContinue: false,
            ...overrides
        }));
        // Setup default body size to 1MB
        this.app.use(body_parser_1.default.json({ limit: 1 * 1024 * 1024 }));
        // Parse incoming request body and append data to `req.body`
        this.app.use(body_parser_1.default.urlencoded({ extended: true }));
        // Setup response timer
        this.app.use((req, res, next) => {
            req['start'] = (0, cm_libs_utils_1.timeNow)();
            return next();
        });
        // Setup error handler
        if (overrides?.errorUrl) {
            common_1.ErrorHandler.init(overrides.errorUrl);
        }
        // Setup authentication bearer token middleware
        if (overrides?.identityUrl) {
            common_1.AuthBearer.init(overrides.identityUrl);
            this.app.use(common_1.authBearerTokenMiddleware);
        }
        // Setup authentication basic token middleware
        if (overrides?.fnVerify) {
            common_1.AuthBasic.init(overrides.fnVerify);
            this.app.use(common_1.authBasicTokenMiddleware);
        }
        // Setup logger middleware
        this.app.use(common_1.logResMiddleware);
        // Setup debug logger middleware
        if (this.devMode && !overrides?.disableDebugTrace) {
            this.app.use(common_1.dumpReqMiddleware);
            this.app.use(common_1.dumpResMiddleware);
            this.app.use(common_1.dumpResErrorMiddleware);
        }
        return this;
    }
    async term() {
        if (this.listener) {
            await new Promise((resolve, reject) => {
                try {
                    this.listener?.close((error) => {
                        /* istanbul ignore next */
                        return error ? reject(error) : resolve();
                    });
                }
                catch (error) {
                    /* istanbul ignore next */
                    return reject(error);
                }
            });
            this.listener = undefined;
        }
    }
    async run(port, hostname) {
        // Setup default endpoints
        this.app.get('/', this.handler(this, this.onRoot, undefined, false));
        this.app.get('/liveness', this.handler(this, this.onLiveness, undefined, false));
        this.app.get('/readiness', this.handler(this, this.onReadiness, undefined, false));
        // Setup API documentation endpoint
        this.app.use('/apidoc', express_1.default.static('apidoc', { index: ['index.html'] }));
        // Setup 404 handler middleware
        this.app.use(common_1.finalMiddleware); // NOTE: Make sure this handler is place at the last of the router stack
        // Setup error middleware
        this.app.use(common_1.errorMiddleware);
        return new Promise((resolve, reject) => {
            try {
                port = port || 8080;
                this.listener = this.app?.listen(port, hostname || '0.0.0.0', () => {
                    return resolve(`http://${hostname || '*'}:${port}`);
                });
            }
            catch (error) {
                /* istanbul ignore next */
                return reject(error);
            }
        });
    }
    handler(thisObject, func, authDirective, logging = true, raw = false) {
        return async (req, res, next) => {
            try {
                if (logging) {
                    if (this._devMode) {
                        (0, cm_libs_context_aware_logger_1.log)(cm_libs_context_aware_logger_1.Level.CRITICAL, () => {
                            const headers = JSON.stringify(req.headers, null, 2);
                            const body = JSON.stringify(req.body, null, 2);
                            return `TRACE REQ: ${req.method} ${req.url}\nHEADERS: ${headers}\nBODY: ${body}`;
                        });
                    }
                    else {
                        (0, cm_libs_context_aware_logger_1.log)(cm_libs_context_aware_logger_1.Level.NORMAL, () => `REQ: ${req.method} ${req.url} HTTP/${req.httpVersion} @${(0, common_1.getIp)(req)}`);
                    }
                }
                else {
                    req['logExcluded'] = true;
                }
                if (authDirective) {
                    const accessToken = await (0, common_1.auth)(req, authDirective);
                    if (authDirective.hasAbility) {
                        try {
                            (0, common_1.hasAbility)(accessToken, authDirective.hasAbility);
                        }
                        catch (error) {
                            const localeCode = req.get('language');
                            const { code, message } = await common_1.ErrorHandler.translate(error, error.statusCode, localeCode);
                            error.code = code;
                            error.message = message;
                            (0, cm_libs_context_aware_error_1.throwFromError)(error, cm_libs_context_aware_error_1.errors.FORBIDDEN, { spec: authDirective.hasAbility });
                        }
                    }
                }
                const data = await func.apply(thisObject, [req, res, next]);
                if (undefined !== data) {
                    (0, common_1.goodRes)(res, data, undefined, raw);
                }
            }
            catch (error) {
                next(error);
            }
        };
    }
    fileHandler(cachePath, maxFileSize, fileFieldName) {
        return async (req, res, next) => {
            try {
                const options = {
                    dest: cachePath,
                    limits: {
                        fileSize: maxFileSize
                    }
                };
                const upload = (0, multer_1.default)(options);
                return upload.single(fileFieldName || 'file')(req, res, next);
            }
            catch (error) {
                /* istanbul ignore next */
                return next(error);
            }
        };
    }
    bootstrap() {
        // Catches CTRL+C signal
        process.on('SIGINT', async () => {
            if (!this._terminated) {
                await this.shutdown();
                this._terminated = true;
            }
        });
        // Listen for TERM signal. E.g. kill
        process.on('SIGTERM', async () => {
            if (!this._terminated) {
                await this.shutdown();
                this._terminated = true;
            }
        });
        // Catches SIGUSR1 (10) signal to increase log severity (circled)
        process.on('SIGUSR1', () => {
            cm_libs_context_aware_logger_1.ContextAwareLogger.instance.level = null;
            (0, cm_libs_context_aware_logger_1.log)(cm_libs_context_aware_logger_1.Level.CRITICAL, () => `Logger level: ${cm_libs_context_aware_logger_1.ContextAwareLogger.instance.level}`);
        });
        // Catches uncaught exceptions
        process.on('uncaughtException', async (error, origin) => {
            if (process.stderr.writable) {
                process.stderr.write(`uncaughtException: ${error}, ${origin}\n`);
            }
        });
        // Catches unhandled rejection
        process.on('unhandledRejection', (reason, stack) => {
            if (process.stderr.writable) {
                process.stderr.write(`unhandledRejection: ${reason}, ${stack}\n`);
            }
        });
        process.env.NODE_ENV = common_1.ConfigBase.get('NODE_ENV', (0, cm_libs_utils_1.allowNullOrEmpty)(cm_libs_utils_1.checkString, 'production'));
        this.devMode = 'dev' === process.env.NODE_ENV;
        // Setup logger
        cm_libs_context_aware_logger_1.ContextAwareLogger.init(common_1.ConfigBase.get('LOG_LEVEL', (0, cm_libs_utils_1.allowNullOrEmpty)((value) => {
            const v = (0, cm_libs_utils_1.checkString)(value);
            return (0, cm_libs_utils_1.isUndefinedOrNull)(cm_libs_context_aware_logger_1.Level[v]) ? cm_libs_context_aware_logger_1.Level.NORMAL : cm_libs_context_aware_logger_1.Level[v];
        })));
    }
    /**
     * @api {get} /liveness Liveness health check
     * @apiVersion 1.0.0
     * @apiName onLiveness
     * @apiGroup HealthCheck
     * @apiPermission none
     *
     * @apiDescription Service liveness health check.
     *
     * @apiSuccessExample Success-Response:
     *     HTTP/1.1 200 OK
     *
     * @apiErrorExample Error-Response:
     *     HTTP/1.1 500 Service is offline
     */
    async onLiveness(req, res) {
        res.status(this.liveness ? 200 : 500).send();
    }
    /**
     * @api {get} /readiness Readiness health check
     * @apiVersion 1.0.0
     * @apiName onReadiness
     * @apiGroup HealthCheck
     * @apiPermission none
     *
     * @apiDescription Service readiness health check.
     *
     * @apiSuccessExample Success-Response:
     *     HTTP/1.1 200 OK
     *
     * @apiErrorExample Error-Response:
     *     HTTP/1.1 429 Service is not ready
     */
    async onReadiness(req, res) {
        res.status(this.readiness ? 200 : 429).send();
    }
    /**
     * @api {get} / Service information
     * @apiVersion 1.0.0
     * @apiName onRoot
     * @apiGroup HealthCheck
     * @apiPermission none
     *
     * @apiDescription Service information.
     *
     * @apiSuccess {String} message Service information
     *
     * @apiSuccessExample Success-Response:
     *     HTTP/1.1 200 OK
     *     {
     *       "message": "Welcome to the coolest API on earth - ServiceName:1.0.0"
     *     }
     */
    onRoot(req, res) {
        res.json({
            message: `Welcome to the coolest API on earth - ${this.name}:${this.version}!`
        });
    }
    get devMode() {
        return this._devMode;
    }
    set devMode(value) {
        this._devMode = value;
    }
}
exports.AppBase = AppBase;
//# sourceMappingURL=app-base.js.map